﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;

namespace CableExperiment
{
    public partial class Page : UserControl
    {
        const double MAX_CONTROL_POINT_Y_OFFSET = 100;
        const double CUTOFF_DIFFERENCE = 20;
        const double FLATTENING_TOLERANCE = 100;
        double cableLength;
        FrameworkElement elementDragged;
        Point plugPointStart;
        Point mousePointStart;
        PathGeometry shadowCableGeometry = new PathGeometry();
        PathFigure shadowCableFigure = new PathFigure();
        BezierSegment shadowCableSegment = new BezierSegment();

        public Page()
        {
            InitializeComponent();

            SetPlugPositions();
            cableLength = GetBezierLength(cableGeometry);
            shadowCableFigure.Segments.Add(shadowCableSegment);
            shadowCableGeometry.Figures.Add(shadowCableFigure);
        }

        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs args)
        {
            if (args.OriginalSource is FrameworkElement &&
                    ((args.OriginalSource as FrameworkElement).Parent == plug1 || 
                     (args.OriginalSource as FrameworkElement).Parent == plug2))
            {
                elementDragged = (args.OriginalSource as FrameworkElement).Parent as FrameworkElement;

                if (elementDragged == plug1)
                    plugPointStart = cableFigure.StartPoint;
                else
                    plugPointStart = cableSegment.Point3;

                // Define points for shadow geometry
                shadowCableFigure.StartPoint = cableFigure.StartPoint;
                shadowCableSegment.Point1 = cableSegment.Point1;
                shadowCableSegment.Point2 = cableSegment.Point2;
                shadowCableSegment.Point3 = cableSegment.Point3;

                mousePointStart = args.GetPosition(this);
                CaptureMouse();
            }
        }

        protected override void OnMouseMove(MouseEventArgs args)
        {
            Point mousePoint = args.GetPosition(this);

            if (elementDragged == null)
                return;

            // Set new points in shadow geometry
            if (elementDragged == plug1)
            {
                shadowCableFigure.StartPoint = new Point(plugPointStart.X + mousePoint.X - mousePointStart.X,
                                                         plugPointStart.Y + mousePoint.Y - mousePointStart.Y);
                shadowCableSegment.Point1 = new Point(shadowCableFigure.StartPoint.X, shadowCableSegment.Point1.Y);
            }

            else if (elementDragged == plug2)
            {
                shadowCableSegment.Point3 = new Point(plugPointStart.X + mousePoint.X - mousePointStart.X,
                                                      plugPointStart.Y + mousePoint.Y - mousePointStart.Y);
                shadowCableSegment.Point2 = new Point(shadowCableSegment.Point3.X, shadowCableSegment.Point2.Y);
            }

            // Get a range of Y's for the Bezier control points
            double yLowerBound = MAX_CONTROL_POINT_Y_OFFSET + Math.Max(shadowCableFigure.StartPoint.Y, shadowCableSegment.Point3.Y);
            double yUpperBound = shadowCableFigure.StartPoint.Y + shadowCableSegment.Point3.Y + cableLength;

            if (yLowerBound > yUpperBound)
                return;

            // Find lengths based on two Y's
            double lowerBoundLength = SetNewControlPointsAndGetLength(yLowerBound);
            double upperBoundLength = SetNewControlPointsAndGetLength(yUpperBound);

            if (lowerBoundLength > cableLength || upperBoundLength < cableLength)
                return;

            // Narrow in on cable length
            double yTrial = (yLowerBound + yUpperBound) / 2;

            while (yUpperBound - yLowerBound > CUTOFF_DIFFERENCE)
            {
                double trialLength = SetNewControlPointsAndGetLength(yTrial);

                if (trialLength < cableLength)
                    yLowerBound = yTrial;
                else
                    yUpperBound = yTrial;

                yTrial = (yLowerBound + yUpperBound) / 2;
            }

            SetNewControlPointsAndGetLength(yTrial);

            // Set cable points to shadow points
            cableFigure.StartPoint = shadowCableFigure.StartPoint;
            cableSegment.Point1 = shadowCableSegment.Point1;
            cableSegment.Point2 = shadowCableSegment.Point2;
            cableSegment.Point3 = shadowCableSegment.Point3;

            SetPlugPositions(); 
        }

        // Mouse up: Release capture
        protected override void OnMouseLeftButtonUp(MouseButtonEventArgs args)
        {
            elementDragged = null;
            ReleaseMouseCapture();
        }
 
        // Calculate length of a the trial geometry based on control-point Y values
        double SetNewControlPointsAndGetLength(double y)
        {
            shadowCableSegment.Point1 = new Point(shadowCableFigure.StartPoint.X, y);
            shadowCableSegment.Point2 = new Point(shadowCableSegment.Point3.X, y);
            return GetBezierLength(shadowCableGeometry);
        }

        double GetVectorLength(Point pt1, Point pt2)
        {
            return Math.Sqrt(Math.Pow(pt1.X - pt2.X, 2) + Math.Pow(pt1.Y - pt2.Y, 2));
        }

        void SetPlugPositions()
        {
            Canvas.SetLeft(plug1, cableFigure.StartPoint.X);
            Canvas.SetTop(plug1, cableFigure.StartPoint.Y);

            Canvas.SetLeft(plug2, cableSegment.Point3.X);
            Canvas.SetTop(plug2, cableSegment.Point3.Y);
        }

        double GetBezierLength(PathGeometry geometry)
        {
            double length = 0;
            PathFigure pathFigure = geometry.Figures[0];
            BezierSegment bezierSegment = pathFigure.Segments[0] as BezierSegment;

            int numPoints = (int)((GetVectorLength(pathFigure.StartPoint, bezierSegment.Point1) +
                                   GetVectorLength(bezierSegment.Point1, bezierSegment.Point2) +
                                   GetVectorLength(bezierSegment.Point2, bezierSegment.Point3)) / FLATTENING_TOLERANCE);

            Point ptPrevious = pathFigure.StartPoint;

            for (int i = 1; i < numPoints; i++)
            {
                double t = i / (numPoints - 1.0);

                double x = (1 - t) * (1 - t) * (1 - t) * pathFigure.StartPoint.X +
                           3 * t * (1 - t) * (1 - t) * bezierSegment.Point1.X +
                           3 * t * t * (1 - t) * bezierSegment.Point2.X +
                           t * t * t * bezierSegment.Point3.X;

                double y = (1 - t) * (1 - t) * (1 - t) * pathFigure.StartPoint.Y +
                           3 * t * (1 - t) * (1 - t) * bezierSegment.Point1.Y +
                           3 * t * t * (1 - t) * bezierSegment.Point2.Y +
                           t * t * t * bezierSegment.Point3.Y;

                Point pt = new Point(x, y);
                length += GetVectorLength(pt, ptPrevious);
            }

            return length;
        }
    }
}
